#pragma once

#include "MemoryBase.h"

#include <ArmarXCore/core/services/tasks/PeriodicTask.h>

namespace armarx::armem::server::ltm
{
    template <class _CoreSegmentT>
    class BufferedMemoryBase : public MemoryBase<_CoreSegmentT>
    {
        using Base = MemoryBase<_CoreSegmentT>;

    public:
        BufferedMemoryBase() :
            Base()
        {
        }

        BufferedMemoryBase(const MemoryID& id) :
            Base(id),
            buffer(id)
        {
        }

        virtual ~BufferedMemoryBase() = default;

        armem::wm::Memory getBuffer() const
        {
            return buffer;
        }

        void directlyStore(const armem::wm::Memory& memory)
        {
            TIMING_START(LTM_Memory_DirectlyStore);
            if (!this->pipeline.memFreqFilter->accept(memory))
            {
                ARMARX_WARNING << deactivateSpam() << "Ignoring to put a Memory into the LTM because it got filtered.";
                return;
            }
            else
            {
                _directlyStore(memory);
            }
            TIMING_END(LTM_Memory_DirectlyStore);
        }

        void storeBuffer()
        {
            std::lock_guard l(bufferMutex);
            if (buffer.empty())
            {
                ARMARX_INFO << deactivateSpam() << "Cannot store an empty buffer. Ignoring.";
                return;
            }

            this->directlyStore(buffer);
            buffer.clear();
        }

        virtual void createPropertyDefinitions(PropertyDefinitionsPtr& defs, const std::string& prefix) override
        {
            Base::createPropertyDefinitions(defs, prefix);

            defs->optional(storeFrequency, prefix + ".buffer.storeFreq", "Frequency to store the buffer to the LTM in Hz.");
        }

    protected:
        virtual void _directlyStore(const armem::wm::Memory& memory) = 0;

        void _store(const armem::wm::Memory& memory) override
        {
            std::lock_guard l(bufferMutex);
            buffer.append(memory);
            if (!task)
            {
                int waitingTimeMs = 1000.f / storeFrequency;
                task = new armarx::PeriodicTask<BufferedMemoryBase>(this, &BufferedMemoryBase::storeBuffer, waitingTimeMs);
                task->start();
            }
        }

    protected:

        /// Internal memory for data consolidated from wm to ltm (buffer)
        /// The to-put-to-ltm buffer (contains data in plain text)
        /// This buffer may still be filtered.
        /// This means that it is not guaranteed that all data in the buffer will be stored in the ltm
        armem::wm::Memory buffer;

        /// The periodic'task to store the content of the buffer to the ltm
        typename armarx::PeriodicTask<BufferedMemoryBase>::pointer_type task = nullptr;

        /// The frequency to store data to the ltm
        float storeFrequency = 10;

        /// a mutex to access the buffer object
        std::mutex bufferMutex;

    };
} // namespace armarx::armem::server::ltm
