/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <typeindex>
#include <typeinfo>
#include <map>
#include <string>

// Simox
#include <SimoxUtility/algorithm/string.h>

// ArmarX
#include <RobotAPI/interface/aron.h>


/**
 * This file contains basic enums to distinguish aron::data and aron::types and utility functions to generate maps converting the neum to strings or vice versa.
 * Further, this file contains a conversion map from ice_ids to descriptors.
 *
 * Every data- or type-object has a descriptor, containing the correct enum for this object. All aron-variants must be either:
 * DATA:
 *  - Dict
 *  - List
 *  - NDArray
 *  - Int
 *  - Long
 *  - Float
 *  - Double
 *  - String
 *  - Bool
 *
 * TYPE:
 *  - Object
 *  - Dict
 *  - List
 *  - Pair
 *  - Tuple
 *  - NDArray
 *  - Matrix
 *  - Orientation
 *  - Position
 *  - Pose
 *  - Quaternion
 *  - Image
 *  - Pointcloud
 *  - Int Enum
 *  - Int
 *  - Long
 *  - Float
 *  - Double
 *  - String
 *  - Bool
 *  - Time
 *
 *  A variant should never have eUnknown as its internal type (The enums are const and only set on construction).
 *  If you want to execute a specific function for a specific variant, please refer to the visitor implementation.
 */
namespace armarx::aron::conversion::util
{
    template <class T1, class T2>
    std::map<T2, T1> InvertMap(const std::map<T1, T2>& m)
    {
        std::map<T2, T1> ret;
        for (const auto &[key, val] : m)
        {
            ret.emplace(val, key);
        }
        return ret;
    }

    template <class T>
    std::map<T, std::string> InvertMap(const std::map<std::string, T>& m)
    {
        return InvertMap<std::string, T>(m);
    }

    template <class T>
    std::map<std::string, T> InvertMap(const std::map<T, std::string>& m)
    {
        return InvertMap<T, std::string>(m);
    }
}

namespace armarx::aron::type
{
    const std::vector<type::Maybe> AllMaybeTypes =
    {
        type::Maybe::eNone,
        type::Maybe::eOptional,
        type::Maybe::eRawPointer,
        type::Maybe::eSharedPointer,
        type::Maybe::eUniquePointer
    };

    enum class Descriptor
    {
        eList,
        eObject,
        eTuple,
        ePair,
        eDict,
        eNDArray,
        eMatrix,
        eQuaternion,
        ePointCloud,
        ePosition,
        eOrientation,
        ePose,
        eImage,
        eIntEnum,
        eInt,
        eLong,
        eFloat,
        eDouble,
        eString,
        eBool,
        eTime,
        eAnyObject,
        eUnknown = -1
    };

    const std::vector<type::Descriptor> AllDescriptors =
    {
        Descriptor::eList,
        Descriptor::eObject,
        Descriptor::eTuple,
        Descriptor::ePair,
        Descriptor::eDict,
        Descriptor::eNDArray,
        Descriptor::eMatrix,
        Descriptor::eQuaternion,
        Descriptor::ePointCloud,
        Descriptor::ePosition,
        Descriptor::eOrientation,
        Descriptor::ePose,
        Descriptor::eImage,
        Descriptor::eIntEnum,
        Descriptor::eInt,
        Descriptor::eLong,
        Descriptor::eFloat,
        Descriptor::eDouble,
        Descriptor::eString,
        Descriptor::eBool,
        Descriptor::eTime,
        Descriptor::eAnyObject,
        Descriptor::eUnknown
    };

    namespace defaultconversion::string
    {
        // Maybe
        const std::map<type::Maybe, std::string> Maybe2String =
        {
            {Maybe::eNone, "armarx::aron::type::Maybe::eNone"},
            {Maybe::eRawPointer, "armarx::aron::type::Maybe::eRawPointer"},
            {Maybe::eSharedPointer, "armarx::aron::type::Maybe::eSharedPointer"},
            {Maybe::eUniquePointer, "armarx::aron::type::Maybe::eUniquePointer"},
            {Maybe::eOptional, "armarx::aron::type::Maybe::eOptional"}
        };

        // Descriptor
        const std::map<type::Descriptor, std::string> Descriptor2String =
        {
            {Descriptor::eList, "armarx::aron::type::Descriptor::eList"},
            {Descriptor::eObject, "armarx::aron::type::Descriptor::eObject"},
            {Descriptor::eDict, "armarx::aron::type::Descriptor::eDict"},
            {Descriptor::ePair, "armarx::aron::type::Descriptor::ePair"},
            {Descriptor::eTuple, "armarx::aron::type::Descriptor::eTuple"},
            {Descriptor::eNDArray, "armarx::aron::type::Descriptor::eNDArray"},
            {Descriptor::eMatrix, "armarx::aron::type::Descriptor::eMatrix"},
            {Descriptor::eOrientation, "armarx::aron::type::Descriptor::eOrientation"},
            {Descriptor::eQuaternion, "armarx::aron::type::Descriptor::eQuaternion"},
            {Descriptor::eImage, "armarx::aron::type::Descriptor::eImage"},
            {Descriptor::ePointCloud, "armarx::aron::type::Descriptor::ePointCloud"},
            {Descriptor::eNDArray, "armarx::aron::type::Descriptor::eNDArray"},
            {Descriptor::ePose, "armarx::aron::type::Descriptor::ePose"},
            {Descriptor::ePosition, "armarx::aron::type::Descriptor::ePosition"},
            {Descriptor::eIntEnum, "armarx::aron::type::Descriptor::eIntEnum"},
            {Descriptor::eInt, "armarx::aron::type::Descriptor::eInt"},
            {Descriptor::eLong, "armarx::aron::type::Descriptor::eLong"},
            {Descriptor::eFloat, "armarx::aron::type::Descriptor::eFloat"},
            {Descriptor::eDouble, "armarx::aron::type::Descriptor::eDouble"},
            {Descriptor::eString, "armarx::aron::type::Descriptor::eString"},
            {Descriptor::eBool, "armarx::aron::type::Descriptor::eBool"},
            {Descriptor::eTime, "armarx::aron::type::Descriptor::eTime"},
            {Descriptor::eAnyObject, "armarx::aron::type::Descriptor::eAnyObject"},
            {Descriptor::eUnknown, "armarx::aron::type::Descriptor::eUnknown"}
        };
    }

    namespace defaultconversion::typeinfo
    {
        // hash type
        const std::map<size_t, Descriptor> TypeId2Descriptor = {
            {typeid(aron::type::dto::List).hash_code(), Descriptor::eList},
            {typeid(aron::type::dto::AronObject).hash_code(), Descriptor::eObject},
            {typeid(aron::type::dto::Tuple).hash_code(), Descriptor::eTuple},
            {typeid(aron::type::dto::Pair).hash_code(), Descriptor::ePair},
            {typeid(aron::type::dto::Dict).hash_code(), Descriptor::eDict},
            {typeid(aron::type::dto::NDArray).hash_code(), Descriptor::eNDArray},
            {typeid(aron::type::dto::Matrix).hash_code(), Descriptor::eMatrix},
            {typeid(aron::type::dto::Quaternion).hash_code(), Descriptor::eQuaternion},
            {typeid(aron::type::dto::PointCloud).hash_code(), Descriptor::ePointCloud},
            {typeid(aron::type::dto::Position).hash_code(), Descriptor::ePosition},
            {typeid(aron::type::dto::Orientation).hash_code(), Descriptor::eOrientation},
            {typeid(aron::type::dto::Pose).hash_code(), Descriptor::ePose},
            {typeid(aron::type::dto::Image).hash_code(), Descriptor::eImage},
            {typeid(aron::type::dto::IntEnum).hash_code(), Descriptor::eIntEnum},
            {typeid(aron::type::dto::AronInt).hash_code(), Descriptor::eInt},
            {typeid(aron::type::dto::AronFloat).hash_code(), Descriptor::eFloat},
            {typeid(aron::type::dto::AronLong).hash_code(), Descriptor::eLong},
            {typeid(aron::type::dto::AronDouble).hash_code(), Descriptor::eDouble},
            {typeid(aron::type::dto::AronString).hash_code(), Descriptor::eString},
            {typeid(aron::type::dto::AronBool).hash_code(), Descriptor::eBool},
            {typeid(aron::type::dto::AronTime).hash_code(), Descriptor::eTime},
            {typeid(aron::type::dto::AnyObject).hash_code(), Descriptor::eAnyObject}
        };
    }

    inline type::Descriptor Aron2Descriptor(const type::dto::GenericType& t)
    {
        return defaultconversion::typeinfo::TypeId2Descriptor.at(typeid(t).hash_code());
    }
}

namespace armarx::aron::data
{
    enum class Descriptor
    {
        eList,
        eDict,
        eNDArray,
        eInt,
        eLong,
        eFloat,
        eDouble,
        eString,
        eBool,
        eUnknown = -1
    };

    const std::vector<data::Descriptor> AllDescriptors =
    {
        Descriptor::eList,
        Descriptor::eDict,
        Descriptor::eNDArray,
        Descriptor::eInt,
        Descriptor::eLong,
        Descriptor::eFloat,
        Descriptor::eDouble,
        Descriptor::eString,
        Descriptor::eBool,
        Descriptor::eUnknown
    };

    namespace defaultconversion::string
    {
        // Descriptor
        const std::map<data::Descriptor, std::string> Descriptor2String =
        {
            {Descriptor::eList, "armarx::aron::data::Descriptor::eList"},
            {Descriptor::eDict, "armarx::aron::data::Descriptor::eDict"},
            {Descriptor::eNDArray, "armarx::aron::data::Descriptor::eNDArray"},
            {Descriptor::eInt, "armarx::aron::data::Descriptor::eInt"},
            {Descriptor::eLong, "armarx::aron::data::Descriptor::eLong"},
            {Descriptor::eFloat, "armarx::aron::data::Descriptor::eFloat"},
            {Descriptor::eDouble, "armarx::aron::data::Descriptor::eDouble"},
            {Descriptor::eString, "armarx::aron::data::Descriptor::eString"},
            {Descriptor::eBool, "armarx::aron::data::Descriptor::eBool"},
            {Descriptor::eUnknown, "armarx::aron::data::Descriptor::eUnknown"}
        };
    }

    namespace defaultconversion::typeinfo
    {
        // TypeID
        const std::map<size_t, Descriptor> TypeId2Descriptor = {
            {typeid(aron::data::dto::List).hash_code(), Descriptor::eList},
            {typeid(aron::data::dto::Dict).hash_code(), Descriptor::eDict},
            {typeid(aron::data::dto::NDArray).hash_code(), Descriptor::eNDArray},
            {typeid(aron::data::dto::AronInt).hash_code(), Descriptor::eInt},
            {typeid(aron::data::dto::AronFloat).hash_code(), Descriptor::eFloat},
            {typeid(aron::data::dto::AronLong).hash_code(), Descriptor::eLong},
            {typeid(aron::data::dto::AronDouble).hash_code(), Descriptor::eDouble},
            {typeid(aron::data::dto::AronString).hash_code(), Descriptor::eString},
            {typeid(aron::data::dto::AronBool).hash_code(), Descriptor::eBool},
        };
    }

    inline data::Descriptor Aron2Descriptor(const data::dto::GenericData& t)
    {
        return defaultconversion::typeinfo::TypeId2Descriptor.at(typeid(t).hash_code());
    }

    namespace defaultconversion
    {
        const std::map<Descriptor, aron::type::Descriptor> Data2TypeDescriptor = {
            {Descriptor::eList, aron::type::Descriptor::eList},
            {Descriptor::eDict, aron::type::Descriptor::eDict},
            {Descriptor::eNDArray, aron::type::Descriptor::eNDArray},
            {Descriptor::eInt, aron::type::Descriptor::eInt},
            {Descriptor::eFloat, aron::type::Descriptor::eFloat},
            {Descriptor::eLong, aron::type::Descriptor::eLong},
            {Descriptor::eDouble, aron::type::Descriptor::eDouble},
            {Descriptor::eString, aron::type::Descriptor::eString},
            {Descriptor::eBool, aron::type::Descriptor::eBool},
            {Descriptor::eUnknown, aron::type::Descriptor::eUnknown}
        };

        const std::map<aron::type::Descriptor, Descriptor> Type2DataDescriptor = {
            {aron::type::Descriptor::eList, Descriptor::eList},
            {aron::type::Descriptor::eObject, Descriptor::eDict},
            {aron::type::Descriptor::eTuple, Descriptor::eList},
            {aron::type::Descriptor::ePair, Descriptor::eList},
            {aron::type::Descriptor::eDict, Descriptor::eDict},
            {aron::type::Descriptor::eNDArray, Descriptor::eNDArray},
            {aron::type::Descriptor::eMatrix, Descriptor::eNDArray},
            {aron::type::Descriptor::eQuaternion, Descriptor::eNDArray},
            {aron::type::Descriptor::ePointCloud, Descriptor::eNDArray},
            {aron::type::Descriptor::ePosition, Descriptor::eNDArray},
            {aron::type::Descriptor::eOrientation, Descriptor::eNDArray},
            {aron::type::Descriptor::ePose, Descriptor::eNDArray},
            {aron::type::Descriptor::eImage, Descriptor::eNDArray},
            {aron::type::Descriptor::eIntEnum, Descriptor::eInt},
            {aron::type::Descriptor::eInt, Descriptor::eInt},
            {aron::type::Descriptor::eLong, Descriptor::eLong},
            {aron::type::Descriptor::eFloat, Descriptor::eFloat},
            {aron::type::Descriptor::eDouble, Descriptor::eDouble},
            {aron::type::Descriptor::eString, Descriptor::eString},
            {aron::type::Descriptor::eBool, Descriptor::eBool},
            {aron::type::Descriptor::eTime, Descriptor::eLong},
            {aron::type::Descriptor::eUnknown,  Descriptor::eUnknown},
        };
    }
}
